function this = tune_groups( this, s )

% this = tune_groups( this, settings )
%
% Recorded properties:
%   cputimer, cputime, cost delta, group count, stop trigger level,
%   pr_Reinit_model, pr_Improve_model, pr_Move_var, pr_Merge, pr_Split
%   share_Reinit_model, share_Improve_model, share_Move_var, share_Merge,
%   share_Split, stationary variable sweeps, iterations
%
% settings, valid fields:
%   .stop_trigger            this trigger stops the tuning process
%   .sweeps_to_stabilize     stationary variable sweeps before stabilization
%   .max_split_depth         max split recursion depth
%   .split_max_ngroups       max number of groups for split to be considered
%   .op_pr                   operation probability adaptation:
%      .lead_in                lead-in phase length
%      .lead_in_pr             lead-in probabilities (1x5 vector)
%      .min_pr                 minimum allowed probabilities (1x5 vector)
%      .cputimer_lowpass       lowpass filter strength for cputimer values
%      .efficiency_lowpass     lowpass filter strength for efficiencies
%   .memcheck_interval       memory consumption check frequency (in iters)
%                            (zero to disable)
%   .Reinit_model            settings for primitive op Reinit_model
%   .Improve_model           settings for primitive op Improve_model
%   .Move_var                settings for primitive op Move_var
%   .Merge                   settings for primitive op Merge
%   .Split                   settings for primitive op Split
%
% Tunes the grouping according to given settings. Primitive operation
% probabilities are automatically adjusted according to results obtained from
% stats measurements. Tuning is stopped when 'stop_trigger' activates.
%
%
% Probability adaptation
%
% All primitive operation statistics are read from the stats object at call
% level. Variables 'cputimer' and 'efficiency' are used: both are filtered
% through a lowpass filter (flt_lowpass), efficiencies are divided by
% cputimer readings, the results are normalized and then used as
% probabilities.
%
% Initially the given 'lead_in_pr' probabilities are used. When 'lead_in'
% calls have been done for an operation, then the automatically
% generated probabilities are used. Between these states the probability
% will be the linear interpolation between these values.
% 
% Also, the probabilities will never go under 'min_pr' values (to avoid
% 'locking out' an operation).


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global STATS State;




% ---- initialization ----


% initialize probability refreshing
pr_state                   = pr_init( s.op_pr );
[pr_state, pr, raw_shares] = pr_refresh( pr_state );

% initialize the Move_var's nontested variable queue
move_var_nontested_queue = State.vars( randperm(length(State.vars)) );

% initialize memory consumption checking
if s.memcheck_interval == 0
  memcheck_lastval = NaN;
else
  memcheck_counter = 0;
  memcheck_lastval = ...
      getfield( whos('State'), 'bytes' ) + getfield( whos('STATS'), 'bytes' );
end


% the algorithm iteration counter
iter = 0;

% counts successive iterations with no changes in the grouping
stationary_var_sweeps = 0;


% copy and enable the main stop trigger
stop_trigger = set_enabled( s.stop_trigger, 1 );




% ---- mainloop ----


% tune until stop trigger activates
STATS = measure( STATS, 'tune_groups', 'start', State, ...
                 [length(State.Ind) 0 memcheck_lastval pr raw_shares ...
                  stationary_var_sweeps iter] );
[stop_trigger, trigger_level] = get_level( stop_trigger );
while trigger_level < 1
  
  
  % ---- update probabilities etc ----
  
  % Refresh the operation probabilities
  [pr_state, pr, raw_shares] = pr_refresh( pr_state );

  % Check if stabilization should be activated
  if stationary_var_sweeps >= s.sweeps_to_stabilize
    % lock out all operations but Move_var
    pr = [0 0 1 0 0];
  end
  
  % Check memory consumption
  if s.memcheck_interval ~= 0
    memcheck_counter = mod( memcheck_counter + 1, s.memcheck_interval );
    if memcheck_counter == 0
      memcheck_lastval = ...
          getfield(whos('State'), 'bytes') + getfield(whos('STATS'), 'bytes');
    end
  end
  
  
  % ---- call primitive operations ----
  
  grouping_changed = false;
  
  % Reinit_model?
  if rand < pr(1)
    % choose a random active group
    group = randperm( length(State.Ind) ); group = group(1);
    ind = State.Ind( group );
    % try to reinit its model
    Reinit_model( ind, s.Reinit_model );
  end
  
  % Improve_model?
  if rand < pr(2)
    % choose a random active group
    group = randperm( length(State.Ind) ); group = group(1);
    ind = State.Ind( group );
    % improve its model
    Improve_model( ind, s.Improve_model );
  end
  
  % Move_var?
  if rand < pr(3)
    % pop next variable from the queue
    var = move_var_nontested_queue(1);
    move_var_nontested_queue(1) = [];
    % try to move the next var
    success = Move_var( var, s.Move_var );
    grouping_changed = grouping_changed | success;
    % increment the stationary variable sweep counter
    stationary_var_sweeps = stationary_var_sweeps + 1 / length(State.vars);
  end
  
  % Merge?
  if rand < pr(4)
    success = Merge( s.Merge );
    grouping_changed = grouping_changed | success;
  end
  
  % Split?
  if rand < pr(5) && s.max_split_depth > 0 && ...
        s.split_max_ngroups >= length(State.Ind)
    success = Split( s.Split, s.max_split_depth );
    grouping_changed = grouping_changed | success;
  end
  
  
  % ---- update some variables ----
  
  % reset the stationary_var_sweeps counter if grouping has changed
  if grouping_changed, stationary_var_sweeps = 0; end
  
  % keep the Move_var's queue nonempty (reset if empty)%or grouping has changed
  if isempty(move_var_nontested_queue)   % || grouping_changed
    move_var_nontested_queue = State.vars( randperm(length(State.vars)) );
  end
  
  % increment the algorithm iteration counter
  iter = iter + 1;
  
  
  % --------
  
  % sanity check
  %sanity_check( this, State, s );
  
  % record progress and read stop trigger level
  STATS = measure( STATS, 'tune_groups', 'checkpoint', State, ...
                   [length(State.Ind) trigger_level memcheck_lastval ...
                    pr raw_shares stationary_var_sweeps iter] );
  [stop_trigger, trigger_level] = get_level( stop_trigger );
  
end
STATS = measure( STATS, 'tune_groups', 'stop', State, ...
                 [length(State.Ind) trigger_level memcheck_lastval ...
                  pr raw_shares stationary_var_sweeps iter] );




% --------


return;








function this = pr_init( s )
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% input filter chain:
%   lowpass(min(0, 'total cost')) / lowpass('cputimer')^2
input = ...
    flt_eval(                                               ...
        ... % Q = dC / dt^2
        '(-in{1}) ./ (in{2} .^ 2 + eps)',                   ...
        flt_lowpass( s.efficiency_lowpass,                  ...
            flt_eval(                                       ...
                ... % clamp to negative cost delta (while keeping NaNs as
                ... % NaNs). this is necessary with simulated annealing.
                'sign(-in{1}) .* min(0, in{1})',            ...
                     flt_stats_reader( {                    ...
                         {'Reinit_model',  'total cost'},   ...
                         {'Improve_model', 'total cost'},   ...
                         {'Move_var',      'total cost'},   ...
                         {'Merge',         'total cost'},   ...
                         {'Split',         'total cost'}},  ...
                                       'calls' ))),         ...
        flt_lowpass( s.cputimer_lowpass,                    ...
                     flt_stats_reader( {                    ...
                         {'Reinit_model',  'cputimer'},     ...
                         {'Improve_model', 'cputimer'},     ...
                         {'Move_var',      'cputimer'},     ...
                         {'Merge',         'cputimer'},     ...
                         {'Split',         'cputimer'}},    ...
                                       'calls' )) );

this.input      = set_enabled( input, 1 );
this.calls      = zeros(1, 5);
this.last_vals  = zeros(1, 5);

this.lead_in    = s.lead_in;
this.lead_in_pr = s.lead_in_pr;
this.min_pr     = s.min_pr;

return;




function [this, pr, raw_shares] = pr_refresh( this )

% 'pr' contains the final probabilities. 'raw_shares' contains the
% probabilities before normalization and lead-in, which is:
%   lowpass(min(0, 'total cost')) / (lowpass('cputimer') ^ 2 + eps)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% read new outputs (raw shares)
[this.input, vals] = pop( this.input ); outs = ~isnan(vals);
while any(outs)
  
  this.calls = this.calls + outs;
  this.last_vals(outs) = vals(outs);
  
  [this.input, vals] = pop( this.input ); outs = ~isnan(vals);
end

% store the raw shares
raw_shares = this.last_vals;

% normalize and apply lead-in
in_pr      = raw_shares / ( norm(raw_shares, inf) + eps );
in_weights = min( 1, this.calls / this.lead_in );
pr         = max( this.min_pr, ...
                  in_weights .* in_pr + (1-in_weights) .* this.lead_in_pr );

return;
