function cs = cs_basic( data, state, varargin )

% cs = cs_basic( data, state, [cs_args] )
%
% Default settings for the ctrl_basic controller. Accepted extra arguments are
% explained in the help section of ivga().
%
% Operation probabilities:
%   - All operations except Split have fixed probabilites. The Split
%     probability is a heuristic based on the average group size and current
%     iteration number.
%   - Split is always disabled if already splitting and all variables are in
%     one group (this is hard-coded into Split()).
%
% Default stopping criterion:
%   Evaluation is stopped when the lowpass-filtered efficiency drops below 0.03
%   times the peak value of the lowpass-filtered efficiency during 5 *
%   variable_count iterations (at least 5 * variable_count iterations are
%   always performed).
%
% Split:
%   - Maximum split recursion depth is limited to 1.
%   - Stopping criterion:
%       Evaluation is stopped if either the group count drops to 1, or 3 *
%       variable_count (local variable count) iterations have elapsed.


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% select the modeler and controller
cs.mdl.class  = 'mdl_vq_real_nominal';
cs.ctrl.class = 'ctrl_basic';




% ---- standard initialization ----


% parse the data dimensions and group setup
[n, dim, setup_mode, setup_params] = ...
    cshelper_parse_grouping_and_dim( cs, data, state, varargin{:} );

% store dimensionality for internal use
cs.private.n = n; cs.private.dim = dim;




% ---- search strategies ----


cs.stl = state_transition_limiter( ...
    'neighborhood_limiter', null_limiter, ...
    'gradient_limiter', greedy_search );




% ---- modeler ----


cs.mdl.max_points       = round(n/2);
cs.mdl.ncentroids_def   = 2;
cs.mdl.iters            = 30;
cs.mdl.check            = 10;




% ---- controller ----


% -- init_groups --

% group setup
cs.ctrl.init_groups.setup_mode    = setup_mode;
cs.ctrl.init_groups.setup_params  = setup_params;

% group model computation
cs.ctrl.init_groups.Reinit_model.stop_trigger    = ...
    trigger_iters( 10, flt_stats_reader('init_groups - Reinit_model','iters'));
cs.ctrl.init_groups.Improve_model.stop_trigger   = ...
    trigger_iters( 0, flt_stats_reader('init_groups - Improve_model','iters'));


% -- tune_groups --

% stopping
if ~isarg('runtime'), varargin = { varargin{:}, 'runtime', 'normal' }; end
cs.ctrl.tune_groups.stop_trigger        = NaN;   % placeholder
cs.ctrl.tune_groups.sweeps_to_stabilize = Inf;   % disabled

% primitive operation probabilities
cs.ctrl.tune_groups.pr_Reinit_model  = 0.3;
cs.ctrl.tune_groups.pr_Improve_model = 0.5;
cs.ctrl.tune_groups.pr_Move_var      = 1;
cs.ctrl.tune_groups.pr_Merge         = 0.25;

% params for the Split probability heuristic
cs.ctrl.tune_groups.Split_min_var_sweeps = 2;
cs.ctrl.tune_groups.Split_invgroupsize_f = 0.2;
cs.ctrl.tune_groups.Split_iter_f         = 0.01;




% ---- primitive operations ----


% Reinit_model
cs.ctrl.tune_groups.Reinit_model.stop_trigger = ...
    trigger_iters( 1, flt_stats_reader('Reinit_model', 'iters') );
% Improve_model
cs.ctrl.tune_groups.Improve_model.stop_trigger = ...
    trigger_iters( 1, flt_stats_reader('Improve_model', 'iters') );
% Move_var
cs.ctrl.tune_groups.Move_var.reinit_lim_trigger = ...
    trigger_iters( 10, flt_stats_reader('Move_var - Reinit_model', 'iters') );
cs.ctrl.tune_groups.Move_var.improve_lim_trigger = ...
    trigger_iters( 0, flt_stats_reader('Move_var - Improve_model', 'iters') );
% Merge
cs.ctrl.tune_groups.Merge.Reinit_model.stop_trigger = ...
    trigger_iters( 1, flt_stats_reader('Merge - Reinit_model', 'iters') );
cs.ctrl.tune_groups.Merge.Improve_model.stop_trigger = ...
    trigger_iters( 0, flt_stats_reader('Merge - Improve_model', 'iters') );
% Split
cs.ctrl.tune_groups.Split.cs                    = @cs_basic_split;
cs.ctrl.tune_groups.Split.current_split_depth   = 0;




% ---- standard finalization ----


% apply common extra arguments
cs = cshelper_apply_common_exargs( cs, data, state, varargin{:} );


return;








function cs = cs_basic_split( data, state, varargin )

% Settings for the recursive Split call.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% inherit from cs_basic
cs  = cs_basic( data, state, varargin{:} );
dim = cs.private.dim;


% -- override some fields (shorten the runtime) --

% init_groups
cs.ctrl.init_groups.Reinit_model.stop_trigger = ...
    trigger_iters(1, flt_stats_reader('init_groups - Reinit_model','iters'));
cs.ctrl.init_groups.Improve_model.stop_trigger = ...
    trigger_iters(1, flt_stats_reader('init_groups - Improve_model','iters'));

% tune_groups - stopping
cs.ctrl.tune_groups.stop_trigger = trigger_or( ...
    trigger_range( ...   % stop immediately if all vars go into one group
        [1 1 1 dim], ...
        flt_stats_reader({'tune_groups', 'group count'}, 'iters')), ...
    trigger_iters( 3 * dim, flt_stats_reader('tune_groups', 'iters') ));
cs.ctrl.tune_groups.sweeps_to_stabilize = Inf;

% Split
cs.ctrl.tune_groups.Split_min_var_sweeps = Inf;   % disable Split

return;
