function retval = info( varargin )

% info( result(s) )
% info( filename )
% info( batch_id, cs )
%
% grouping = info( ... )
%
% Displays some general information about the given results. All values are
% averages from the given results.
%
% The first form takes either a single result struct or an n-level cell array
% of result structs obtained from ivga() calls. If a cell array is given, then
% it is flattened and all found result structs are used as input. The second
% form takes a filename and loads an n-level cell array named
% 'all_results'. The third form takes a batch run id and a cs name. The
% all_results files from that batch run must be in current directory. (Note:
% older versions of batch_run did filter these files automatically through this
% function and thus didn't leave the all_results files behind. in that case the
% info() summary can be found from the correspondingly named files with the
% _info suffix.)
%
% A return value can be requested by assignment. This causes the grouping with
% the lowest cost to be returned as a cell array, so that retval{i} contains
% the indices of the variables in group i as an integer vector.
%
% There is a global option VARIABLE_NAMES in IVGA_CONFIG, which can be used
% to give names to the variables when printing the best grouping:
%   global IVGA_CONFIG; IVGA_CONFIG.VARIABLE_NAMES = {'name1', 'name2', ...};
% See the file ivga_config.txt for details.


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.


% A batch run results file could (maybe) also be used as data source..

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

global IVGA_CONFIG;
if isfield( IVGA_CONFIG, 'VARIABLE_NAMES' )
  VARIABLE_NAMES = IVGA_CONFIG.VARIABLE_NAMES;
else VARIABLE_NAMES = []; end


% detect call syntax and fetch the input data
if nargin == 1 & ~isa(varargin{1}, 'char')
  % form 1:
  all_results = varargin{1};
elseif nargin == 1 & isa(varargin{1}, 'char')
  % form 2: load the file
  fprintf( 'Loading data..' );
  wstate = warning('off');
  load( varargin{1}, 'all_results' );
  warning(wstate);
  fprintf( 'OK\n' );
elseif nargin == 2
  % form 3: load the file
  fprintf( 'Loading data..' );
  wstate = warning('off');
  load( ['run_' varargin{1} '_' varargin{2} '_all_results.mat'], ...
        'all_results' );
  warning(wstate);
  fprintf( 'OK\n' );
end


fprintf( 'Processing..' );

% flatten States
States = gather_states( all_results );
stats  = [States.stats];

% prepare some variables
runs          = length(States);
main_ctrl     = get_main_controller_name( stats(1) );
op_names      = get_op_names( stats(1) );
main_ctrl_ind = find( strcmp(op_names, main_ctrl) );
op_order      = [ main_ctrl_ind:length(op_names), 1:main_ctrl_ind-1 ];

op_marks = repmat( '|', 1, length(op_names) );
tmp      = strvcat( op_names{:} );
op_marks( tmp(:,1) == upper(tmp(:,1)) ) = '+';

best_state_ind      = find_best_state( States );
best_state_cost     = total_cost( States(best_state_ind) );
best_state_runtime  = ...
    get_res( stats(best_state_ind), main_ctrl, {'cputimer'}, 'calls' );
best_state_grouping = ...
    { States(best_state_ind).Groups( States(best_state_ind).Ind ).Vars };

if nargout == 1, retval = best_state_grouping; end

fprintf( 'OK\n\n\n' );


fprintf( 'Runs: %d\n', runs );
fprintf( 'Average number of groups: %g\n', length([States.Ind]) / runs );
cost_sum = 0;
runtime_sum = 0;
for run=1:runs
  cost_sum = ...
      cost_sum + get_res( stats(run), main_ctrl, {'total cost'}, 'calls' );
  runtime_sum = ...
      runtime_sum + get_res( stats(run), main_ctrl, {'cputimer'}, 'calls' );
end
fprintf( 'Average cost: %g\n', cost_sum/runs );
fprintf( ['Average runtime: %g s (CPU time)\n'], runtime_sum/runs );

fprintf( '\n' );

if isempty(VARIABLE_NAMES)
  var_field_width = 1 + floor( log10( max( cell2mat(best_state_grouping) ) ) );
else
  var_field_width = size( strvcat(VARIABLE_NAMES{:}), 2 );
end
fprintf( 'Best grouping:\n' );
fprintf( '  Number of groups: %g\n', length(best_state_grouping) );
fprintf( '  Cost: %g\n', best_state_cost );
fprintf( '  Runtime: %g s (CPU time)\n', best_state_runtime );
fprintf( '\n' );
fprintf( '   group |   variables\n' );
fprintf( '  -------|----------------------------------------------------------------------\n' );
for i=1:length(best_state_grouping)
  fprintf( '   %5d |', i );
  if isempty(VARIABLE_NAMES)
    fprintf( [' %' num2str(var_field_width) 'd'], best_state_grouping{i} );
  else
    fprintf( [' %' num2str(var_field_width) 's'], ...
             VARIABLE_NAMES{ best_state_grouping{i} } );
  end
  fprintf( '\n' );
end

fprintf( '\n\n' );

w = warning; warning off;
for op=1:length(op_names)
  run_cputime = 0;
  run_costd   = 0;
  run_calls   = 0;
  run_iters   = 0;
  for run=1:runs
    res = get_res( stats(run), op_names{op}, ...
                               {'cputimer', 'total cost'}, 'calls' );
    res2 = get_res( stats(run), op_names{op}, [], 'iters' );
    if ~isempty(res)
      row = sum( res, 1 );
      run_cputime = run_cputime + (1/runs) * row(1);
      run_costd   = run_costd   + (1/runs) * row(2);
      run_calls   = run_calls   + (1/runs) * size( res, 1 );
      run_iters   = run_iters   + (1/runs) * (size(res2, 1) - size(res, 1));
    end
  end
  table(op,:) = ...
      [ run_cputime, run_cputime / run_calls, run_costd / run_calls, ...
        run_calls, run_iters / run_calls , run_costd, ...
        -(run_costd / run_cputime) ];
end
warning(w);

fprintf( 'Average results for each operation:\n' );
fprintf( '   %2s | %-28s | %15s | %15s | %15s\n', ...
         'op', 'operation name', 'cputime/run', 'cputime/call', ...
         'cost delta / iter' );
fprintf( '  ----|------------------------------|-----------------|' );
fprintf( '-----------------|-----------------\n' );
for op=op_order
  fprintf( '   %2d %s %-28s %s %15.4f | %15.4f | %15.4f\n', ...
           op, op_marks(op), op_names{op}, op_marks(op), ...
           table(op,1), table(op,2), table(op,3) );
end
fprintf( '\n' );

fprintf( '   %2s | %-28s | %15s | %15s | %15s\n', ...
         'op', 'operation name', 'calls/run', 'iters/call', ...
         'cost delta / run' );
fprintf( '  ----|------------------------------|-----------------|' );
fprintf( '-----------------|-----------------\n' );
for op=op_order
  fprintf( '   %2d %s %-28s %s %15.4f | %15.4f | %15.4f\n', ...
           op, op_marks(op), op_names{op}, op_marks(op), ...
           table(op,4), table(op,5), table(op,6) );
end
fprintf( '\n' );

fprintf( '   %2s | %-28s | %15s | %15s | %15s\n', ...
         'op', 'operation name', 'efficiency', '', '' );
fprintf( '  ----|------------------------------|-----------------|' );
fprintf( '-----------------|-----------------\n' );
for op=op_order
  fprintf( '   %2d %s %-28s %s %15.4f | %15.4s | %15.4s\n', ...
           op, op_marks(op), op_names{op}, op_marks(op), ...
           table(op,7), '', '' );
end
fprintf( '\n' );


fprintf( 'Measured variables:\n' );
for op=op_order
  
  var_names = get_var_names( stats(1), op_names{op} );
  fprintf( '  %-28s: ', op_names{op} );
  line_len  = 32;
  for var=1:length( var_names )
    fprintf( '%s', var_names{var} );
    line_len = line_len + length( var_names{var} );
    if var < length( var_names ),
      fprintf( ', ' );
      line_len = line_len + 2;
      if line_len + length( var_names{var+1} ) + 2 > 90
        fprintf(['\n' repmat(' ',1,32)]);
        line_len = 32;
      end
    end
  end
  fprintf( '\n' );
  
end


return;








function all_states = gather_states( all_results )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if isa( all_results, 'struct' )
  
  if isfield( all_results, 'internal' )
    all_states = all_results.internal;
  else
    all_states = all_results;   % is assumed to be already an ivga state
  end
  
else
  
  all_states = [];
  for i=1:length(all_results)
    all_states = cat( 2, all_states, gather_states( all_results{i} ) );
  end
  
end

return;








function ind = find_best_state( States )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


for i=1:length(States)
  cost(i) = total_cost( States(i) );
end

[dummy, ind] = min( cost );

return;
